package iaa.map.client.iseephone;

import java.util.*;
import java.io.*;
import java.text.DecimalFormat;

import iaa.map.GUI.*;
import iaa.map.server.*;

/**
 * Given a path, produces some speakable or text-printable translations.
 * 
 */
public class ReadiblePath 
{
    List rawPath = null;

    List sections = null;

    Runtime rt = null;

    int outputIndex = 0;

    private static DecimalFormat df1dp;
    static {
	df1dp = new DecimalFormat();
	df1dp.setMaximumFractionDigits(1);
    }

    public static class PathSection {
	List segments = null;
	String name = null;
	String action = null;

	double distance;

	public PathSection( String name ) {
	    this.name = name;
	}

	public double getLength()
	{
	    if (segments == null) return 0.0;

	    double total = 0;
	    for (Iterator it = segments.iterator(); it.hasNext(); ) {
		PathSegment ps = (PathSegment)it.next();
		if (ps != null) {
		    total += ps.getLength();
		}
	    }
	    return total;
	}
    }

    public static class JoinedSegments {
	LatLonPosition start1;
	LatLonPosition end1;
	LatLonPosition start2;
	LatLonPosition end2;

	double dist;
    }

    public ReadiblePath( List path ) 
    {
	rawPath = path;

	//	System.out.println("rpath ctor()");

	translate();
    }

    /** Check if PathSegments join each other at at least one end, with maxDist (km) representing
     *  the maximum distance that two ends can be from each other and still considered "meeting".
     *  If so, return a JoinedSegments where the path of travel is from start1->end1->start2->end2.
     *  If not, return null;
     */
    public static JoinedSegments getJoinedSegments( PathSegment ps1, PathSegment ps2, double maxDist )
    {
	JoinedSegments js = null;

	if (SpatialUtil.near(ps1.getStart(), ps2.getStart(), maxDist)) {
	    js = new JoinedSegments();
	    js.start1 = ps1.getEnd();
	    js.end1 = ps1.getStart();
	    js.start2 = ps2.getStart();
	    js.end2 = ps2.getEnd();
	} else if (SpatialUtil.near(ps1.getStart(), ps2.getEnd(), maxDist)) {
	    js = new JoinedSegments();
	    js.start1 = ps1.getEnd();
	    js.end1 = ps1.getStart();
	    js.start2 = ps2.getEnd();
	    js.end2 = ps2.getStart();
	} else if (SpatialUtil.near(ps1.getEnd(), ps2.getStart(), maxDist)) {
	    js = new JoinedSegments();
	    js.start1 = ps1.getStart();
	    js.end1 = ps1.getEnd();
	    js.start2 = ps2.getStart();
	    js.end2 = ps2.getEnd();
	} else if (SpatialUtil.near(ps1.getEnd(), ps2.getEnd(), maxDist)) {
	    js = new JoinedSegments();
	    js.start1 = ps1.getStart();
	    js.end1 = ps1.getEnd();
	    js.start2 = ps2.getEnd();
	    js.end2 = ps2.getStart();
	}
	if (js != null) {
	    js.dist = SpatialUtil.greatCircleDist(js.end1, js.start2);
	}

	return js;
    }

    private JoinedSegments getBestJoinedSegments( PathSection s1, PathSection s2, double maxDist )
    {
	// wonderful n^2 algorithm for figuring out which segments join each other in each section.
	// fortunately n should be very small in most cases.
	JoinedSegments bestJS = null;
	for (Iterator it1 = s1.segments.iterator(); it1.hasNext(); ) {
	    PathSegment ps1 = (PathSegment)it1.next();
	    for (Iterator it2 = s2.segments.iterator(); it2.hasNext(); ) {
		PathSegment ps2 = (PathSegment)it2.next();
		// find how ps1 and ps2 join, if at all
		JoinedSegments js = getJoinedSegments(ps1, ps2, 0.010);
		if (js != null) {
		    if (bestJS == null || bestJS.dist > js.dist) {
			bestJS = js;
		    }
		}
	    }
	}

	return bestJS;
    }

    private String getStartAction( PathSection s1, PathSection s2 )
    {
	JoinedSegments js = getBestJoinedSegments(s1,s2,0.010);

	double az1 = SpatialUtil.sphericalAzimuth(js.start1, js.end1);

	return "Begin by travelling " + SpatialUtil.bearingToDirectionString(az1) + " on";
    }

    /** returns (for now) a text description of the action required to get from
     *  PathSection s1 to s2, assuming they are connected, in the form ("turn left onto...")
     */
    private String getAction( PathSection s1, PathSection s2 ) 
    {
	JoinedSegments js = getBestJoinedSegments(s1,s2,0.010);

	// can't find where these two sections meet up... so leave it to the user??
	if (js == null) {
	    System.err.println("WARNING! ReadiblePath can't find where two pathsegments meet.");
	    return "Then find a way to get to ";
	} else {
	    
	    // System.out.println("dist=" + js.dist);
	    // System.out.println("dist1=" + SpatialUtil.greatCircleDist(js.start1, js.end1) + ", " + js.start1 + ", " + js.end1);
	    // System.out.println("dist2=" + SpatialUtil.greatCircleDist(js.start2, js.end2) + ", " + js.start2 + ", " + js.end2);
	    
	}

	// now get angles of the two adjoining path segments
	double az1 = SpatialUtil.sphericalAzimuth(js.start1, js.end1);
	double az2 = SpatialUtil.sphericalAzimuth(js.start2, js.end2);
	double azDiff = az2 - az1;
	int turn = 0;

	if (azDiff > 0) {
	    if (azDiff < 180.0) {
		turn = 1;
	    } else {
		turn = -1;
	    }
	} else {
	    if (azDiff > -180.0) {
		turn = -1;
	    } else {
		turn = 1;
	    }
	}

	//	System.out.println("az1=" + (int)az1 + " az2=" + (int)az2 + " turn=" + turn);

	switch (turn) {
	case -1:
	    return "Turn left on ";
	case 1:
	    return "Turn right on ";
	default:
	    return null;
	}
    }

    private void translate()
    {
	sections = new ArrayList();
	if (rawPath == null) return;

	PathSection section = null;
	for (Iterator it = rawPath.iterator(); it.hasNext(); ) {
	    PathSegment segment = (PathSegment)it.next();

	    String segName = segment.getStreetName();

	    if (section == null || !section.name.equals(segName)) {
		if (section != null) sections.add(section);
		section = new PathSection(segName);
	    }
	    if (section.segments == null) section.segments = new ArrayList();
	    section.segments.add(segment);
	}
	if (section != null) sections.add(section);

	// now determine what action is required between each section
	if (sections.size() > 1) {
	    for (int i = 0; i < sections.size(); i++) {
		section = (PathSection)sections.get(i);
		
		if (i == 0) {
		    PathSection nextSection = (PathSection)sections.get(i+1);		    
		    section.action = getStartAction(section, nextSection);
		} else {
		    PathSection lastSection = (PathSection)sections.get(i-1);		    
		    section.action = getAction(lastSection, section);
		}
	    }
	} else if (sections.size() == 1) {
	    // there is only one section
	    section = (PathSection)sections.get(0);
	    section.action = "Travel on ";
	}

    }

    private static String cutWSFromEnds( String s )
    {
	int start = 0;
	int end = s.length();
	// get start position
	for (int i = 0; i < s.length(); i++) {
	    char c = s.charAt(i);
	    if (Character.isWhitespace(c)) {
		start = i+1;	       
	    } else {
		break;
	    }
	}
	// get end position
	for (int i = s.length()-1; i >= 0; i--) {
	    char c = s.charAt(i);
	    if (Character.isWhitespace(c)) {
		end = i;
	    } else {
		break;
	    }
	}

	if (start >= end) return "";

	return s.substring(start,end);
    }

    public static void replaceAtEnd( StringBuffer b, String match, String replacement, boolean checkAbbrev )
    {
	if (b == null) return;
	if (b.toString().endsWith(match)) {	    
	    b.delete(b.length() - match.length(), b.length());
	    b.append(replacement);
	} else if (checkAbbrev) {
	    if (b.toString().endsWith(match + ".")) {	    
		b.delete(b.length() - match.length() - 1, b.length());
		b.append(replacement);
	    }
	}
    }

    public static void replaceAtStart( StringBuffer b, String match, String replacement, boolean checkAbbrev )
    {
	if (b == null) return;
	if (b.toString().startsWith(match)) {
	    b.delete(0, match.length());
	    b.insert(0, replacement);
	} else if (checkAbbrev) {
	    if (b.toString().startsWith(match + ".")) {
		b.delete(0, match.length() + 1);
		b.insert(0, replacement);
	    }
	}
    }

    public static void replaceAnywhere( StringBuffer b, String match, String replacement, boolean checkAbbrev )
    {
	if (b == null) return;
	int loc = b.toString().indexOf(match);
	if (loc >= 0) {
	    b.delete(loc, loc + match.length());
	    b.insert(loc, replacement);
	} else if (checkAbbrev) {
	    loc = b.toString().indexOf(match + ".");
	    if (loc >= 0) {
		b.delete(loc, match.length() + 1);
		b.insert(loc, replacement);
	    }
	}
    }

    public static String formatStreetName( String name )
    {
	// rules:
	//  remove ws from start and end
	//  convert "st" at end of name to "street"
	//          "ave" -> "avenue"
	//          "pl" -> "place"
	//          "blvd" -> "boulevard"
	//          "appr" -> "approach"
	//          "br" -> bridge
	//  convert "n/s/e/w" at start to "north/south/..."

	// remove WS from start and end
	StringBuffer result = new StringBuffer(cutWSFromEnds(name.toLowerCase()));

	// translate "st", etc
	replaceAtEnd(result," st", " street", true);
	replaceAtEnd(result," av", " avenue", true);
	replaceAtEnd(result," ave"," avenue", true);
	replaceAtEnd(result," pl", " place", true);
	replaceAtEnd(result," blvd", " boulevard", true);
	replaceAtEnd(result," appr", " approach", true);
	replaceAtEnd(result," br"," bridge", true);

	replaceAtStart(result,"e ","east ", true);	
	replaceAtStart(result,"s ","south ", true);	
	replaceAtStart(result,"w ","west ", true);	
	replaceAtStart(result,"n ","north ", true);	

	replaceAnywhere(result,"houston ", "howston ", false);

	return result.toString();
    }

    public String toString() 
    {
	StringBuffer result = new StringBuffer();
	double totalDist = 0;

	for (Iterator it = sections.iterator(); it.hasNext(); ) {
	    PathSection sec = (PathSection)it.next();

	    String distString = null;
	    double dist = sec.getLength();
	    totalDist += dist;
	    if (dist > 0.150) {	    
		// if dist is > 150m give dist in km with 1 decimal place
		distString = df1dp.format(sec.getLength()) + " kilometers";
	    } else {
		// if dist is <= 150m round to nearest 10m and give in meters
		int distM = (int)(dist * 1000);
		distM = ((int)(distM / 10.0)) * 10;
		if (distM <= 0) distM = 10;
		distString = distM + " meters";
	    }

	    String streetName = formatStreetName(sec.name);

	    result.append(sec.action + " " + streetName + ".\n");
	    result.append("\tTravel " + distString + ".\n");
	}

	String totalDistString = df1dp.format(totalDist) + " kilometers";
	result.append("End of path. Total distance is " + totalDistString + ".");

	return result.toString();
    }

    public static void main( String[] args )
    {
	System.out.println("\"" + args[0] + "\" -> \"" + formatStreetName(args[0]) + "\"");
    }
}
