package iaa.map.client.iseephone;

import java.util.*;
import java.io.*;
import javax.comm.*;

// you need to have an implementation of the javax commAPI for your system for this to work.
// for linux check out: 
//        http://www.interstice.com/~kevinh/linuxcomm.html
// for solaris/windows check out: 
//        http://java.sun.com/products/javacomm/

/**
 * voice modem interface. yeah.
 *
 * much borrowed and mangled from the sun commAPI "SerialDemo" example. 
 */
public class VoiceModemInterface
    implements SerialPortEventListener
{
    public static final int baudrate = 38400; //57600;
    public static final int databits = SerialPort.DATABITS_8;
    public static final int stopbits = SerialPort.STOPBITS_1;
    public static final int parity   = SerialPort.PARITY_NONE;
    public static final int flowctl  = SerialPort.FLOWCONTROL_NONE;

    public static final String DEFAULT_PORT_NAME = "/dev/ttyS1";

    public static final byte DLE_BYTE = 16;

    private CommPortIdentifier portId;
    private SerialPort sPort;

    private OutputStream os;
    private InputStream	is;

    private String portName = DEFAULT_PORT_NAME;

    private boolean open = false;

    private StringBuffer escInputBuffer = new StringBuffer();
    private StringBuffer inputBuffer    = new StringBuffer();
    private List inputList = new ArrayList();

    private boolean expectingEscValue = false;

    public static final int RMD_BUFSIZE = 2000;

    public static class RMDFile {
	FileInputStream fis;
	private byte[] buf;
	
	public RMDFile( String filename, int bufSize ) throws IOException
	{
	    if (bufSize < 2) {
		throw new IllegalArgumentException("bufSize argument must be 2 or greater");
	    }

	    fis = new FileInputStream( filename );
	    buf = new byte[bufSize];

	    // strip off 32byte header
	    if (fis.available() >= 32) {
		byte[] tmp = new byte[32];
		fis.read(tmp,0,32);
	    }
	}

	public int read(byte[] result) throws IOException
	{
	    // the trick is that the modem escape char is \020, so any such 
	    // instances have to be escaped with a second \020

	    // read in an amount of data that is half the allocated buffer size, then
	    // copy the data to a new buffer, doubling and \020 instances

	    if (result.length < buf.length) {
		String msg = "the buffer provided to RMDFile.read() must be at least " 
		    + "as large as the bufSize argument passed to the constructor.";
		    
		throw new IllegalArgumentException(msg);
	    }

	    int biteSize = (int) (buf.length / 2);
	    int n = fis.available();
	    if (n <= 0) {
		return -1;
	    }
	    if (n > biteSize) n = biteSize;

	    // take a small bite
	    int l = fis.read(buf, 0, n);

	    // copy to second buffer
	    int i = 0;
	    int j = 0;
	    while ( i < l ) {
		result[j] = buf[i++];
		if (result[j++] == DLE_BYTE) result[j++] = DLE_BYTE;
	    }

	    return j;
	}

	public int available() throws IOException {
	    return fis.available();
	}

	public void close() throws IOException {
	    fis.close();
	}
    }

    public VoiceModemInterface() 
    {	
    }

    public VoiceModemInterface( String portName )
    {
	this.portName = portName;
    }

    public void open()
    {
	try {
	    portId = CommPortIdentifier.getPortIdentifier(portName);
	} catch (NoSuchPortException e) {	
	    System.err.println("serial port: " + portName + " not available: " + e);
	    return;
	} 

	try {
	    sPort = (SerialPort) portId.open("VoiceModemInterface", 10000);
	} catch (PortInUseException e) {	
	    System.err.println("serial port: " + portName + " in use: " + e);
	    return;
	} 

	setPortParameters();

	try {
	    os = sPort.getOutputStream();
	    is = sPort.getInputStream();
	} catch (IOException e) {
	    sPort.close();
	    System.err.println("could not open serial io streams: " + e);
	    return;
	} 

	// Add this object as an event listener for the serial port.
	try {
	    sPort.addEventListener(this);
	} catch (TooManyListenersException e) {
	    sPort.close();
	    System.err.println("too many serial port listeners: " + e);
	    return;	    
	} 
	
	// Set notifyOnDataAvailable to true to allow event driven input.
	sPort.notifyOnDataAvailable(true);
	
	// Set notifyOnBreakInterrup to allow event driven break handling.
	sPort.notifyOnBreakInterrupt(false);

	// Set receive timeout to allow breaking out of polling loop during
	// input handling.
	try {
	    sPort.enableReceiveTimeout(30);
	} catch (UnsupportedCommOperationException e) {}

	open = true;
    }

    public void close()
    {
	if (!open) return;

	if (sPort != null) {
	    try {
		os.close();
		is.close();
	    } catch (IOException e) {
		System.err.println(e);
	    } 

	    // Close the port.
	    sPort.close();

	    // Remove the ownership listener.
	    //	    portId.removePortOwnershipListener(this);
	} 

	open = false;
    }

    public void setPortParameters() {
	try {
	    sPort.setSerialPortParams(baudrate,
				      databits,
				      stopbits, 
				      parity);
	} catch (UnsupportedCommOperationException e) {
	    System.err.println("could not set serial port params: " + e);
	    return;
	} 

	// Set flow control.
	try {
	    sPort.setFlowControlMode(flowctl);
	} catch (UnsupportedCommOperationException e) {
	    System.err.println("could not set serial port flow ctl: " + e);
	    return;
	} 
    } 

    public void println( String msg ) 
    {
	String out = msg + "\r";
	write(out.getBytes());
    }
    
    public void write( byte[] msg ) 
    {
	write(msg, 0, msg.length);
    }

    public void write( byte[] msg, int offset, int len ) 
    {
	try {
	    os.write(msg, offset, len);
	} catch (IOException e) {
	    System.err.println("OutputStream write error: " + e);
	} 
    }

    public String stripWS( String in ) 
    {
	StringBuffer sb = new StringBuffer();
	for (int i = 0; i < in.length(); i++) {
	    char c = in.charAt(i);
	    if (!Character.isWhitespace(c)) sb.append(c);
	}
	return sb.toString();
    }

    public boolean isExpectedResult( String cmd, String expectedResult, String result, boolean verbose )
    {
	if (expectedResult == null) return true;
	if (result == null) {
	    if (verbose)
		System.err.println("sendCommand: sent \"" + cmd + 
				   "\",\n expected \"" + expectedResult + 
				   "\",\n got nothing.");
	    return false;
	}
	
	expectedResult = stripWS(cmd + expectedResult);
	//expectedResult = cmd + "\n" + expectedResult + "\n";
	if (stripWS(result).startsWith(expectedResult)) {
	    return true;
	}

	if (verbose)
	    System.err.println("sendCommand: sent \"" + cmd + 
			       "\",\n expected \"" + expectedResult + 
			       "\",\n got      \"" + result + "\"");
	
	return false;
    }

    public boolean sendCommand( String cmd, String expectedResult )
    {
	clearInputBuffer();
	println(cmd);

	byte[] buf = new byte[128];
	boolean done = false;
	long timeout = 8000;
	long startTime = System.currentTimeMillis();
	String result = null;
	try {
	    while (!done && (System.currentTimeMillis() - startTime) < timeout) {
		Thread.sleep(100);
		result = getInputBuffer();
		if (isExpectedResult(cmd, expectedResult, result, false)) done = true;
	    }	    
	} catch (InterruptedException e) {
	    System.err.println("INTERRUPTED: " + e);
	}
	
	return !isExpectedResult(cmd, expectedResult, result, true);
    }

    public boolean initModem()
    {
	endTransmitMode();
	if (
	    sendCommand("ATH","OK")             ||
	    sendCommand("ATZ","OK")             ||
	    sendCommand("AT&C1&D2&K3M2L3","OK") ||
	    sendCommand("AT+FCLASS=8","OK")     ||
	    sendCommand("AT+VIP","OK")          ||

	    sendCommand("AT+VSM=141,7200,0,0","OK") ||
	    sendCommand("AT+VPR=0","OK")            ||

	    sendCommand("AT+VLS=4","OK")        ||
	    sendCommand("ATL2","OK")
	    )
	    return true;

	return false;
    }

    public String waitForRing() 
    {
	if (sendCommand("AT+VLS=0","OK") ||
	    sendCommand("AT+FCLASS=8","OK") )
	    return null;

	clearEscInputBuffer();
	boolean gotRing = false;
	String result = null;
	while (!gotRing) {
	    result = getEscInputBuffer();
	    if (result != null) {
		if (result.indexOf("R") >= 0) gotRing = true;
	    }

	    try {
		Thread.sleep(500);
	    } catch (InterruptedException e) {
		System.err.println("INTERRUPTED: " + e);
	    }
	}
	
	return result;
    }

    public boolean startTransmitMode()
    {

	if (sendCommand("AT+VLS=3","OK") ||
	    sendCommand("AT+VTX","CONNECT") )
	    return true;

	return false;
    }

    public boolean endTransmitMode()
    {

	if (sendCommand("\020\003",null)) return true;

	return false;
    }

    /* replaces with get(Esc)InputBuffer()
    public String read()
    {
	int n,l;
	byte[] in = null;
	try {
	    n = is.available();
	    if (n <= 0) return null;
	    in = new byte[n];
	    l = is.read(in,0,n);
	} catch (IOException e) {
	    System.err.println("error reading serial port");
	    return null;
	}
	if (in == null || n < 0 || l < 0) return null;

	return new String(in,0,l);	    
    }
    */

    public boolean transmitFile( String filename )
    {
	return transmitFile( filename, true );
    }
    public boolean transmitFile( String filename, boolean blockUntilDone )
    {
	byte[] buf = new byte[RMD_BUFSIZE];

	List dtmf = new ArrayList();

	clearEscInputBuffer();

	try {

	    RMDFile rmd = new RMDFile(filename, RMD_BUFSIZE);
	    
	    System.out.print("playing " + filename);
	    boolean donePlayback = false;
	    while (!donePlayback) {
		if (rmd.available() > 0) {
		    System.out.print(".");
		    int len = rmd.read(buf);
		    write(buf, 0, len);
		    Thread.sleep(100);
		} else if (blockUntilDone) {
		    Thread.sleep(200);
		    String ein = getEscInputBuffer();
		    if (ein != null) {
			if (ein.lastIndexOf("u") >= 0) {
			    donePlayback = true;
			    System.out.println("\ndone playback");
			}
		    }
		} else {
		    donePlayback = true;
		    System.out.println("\nexiting during playback");
		}
	    }
	    rmd.close();

	} catch (IOException e) {
	    System.err.println("IO error during transmitFile: " + e);
	    return true;
	} catch (InterruptedException e) {
	    
	}
	
	return false;
    }

    public boolean hangup()
    {
	if (sendCommand("ATH","OK")) return true;

	return false;
    }

    public String waitForKeys( int n, int timeout )
    {
	System.out.print("waiting for " + n + " keypress"); 
	System.out.println((n>1)?"es":"");

	clearEscInputBuffer();
	String keys = null;
	String lastKeys = null;
	
	boolean done = false;
	long startTime = System.currentTimeMillis();
	while (!done && (System.currentTimeMillis() - startTime) < timeout) {
	    try {
		Thread.sleep(500);
	    } catch (InterruptedException e) {
		System.err.println("INTERRUPTED: " + e);
	    }

	    keys = interpretKeys(getEscInputBuffer());
	    if (keys != null && keys.length() >= n) done = true;	    
	    
	    if (keys != null && !keys.equals(lastKeys)) {
		System.out.println("[" + keys + "]");
	    }
	    lastKeys = keys;
	}

	return keys;
    }

    private static String interpretKeys( String s )
    {
	StringBuffer keys = new StringBuffer();

	char currentKey = 0;
	for (int i = 0; i < s.length(); i++) {
	    char c = s.charAt(i);
	    if (c == '*' || c == '#' || (c >= '0' && c <= '9')) {
		if (c != currentKey) keys.append(c);
		currentKey = c;
	    } else if (c == '~' || c == '/') {
		currentKey = 0;
	    }
	}

	return keys.toString();
    }
    
    public static void interpretInput( String in )
    {
	System.out.println("keys entered:");
	if (in == null) return;

	char currentKey = 0;
	int i = 0;
	while (i < in.length()) {
	    char c = in.charAt(i);
	    if (c == DLE_BYTE && i < (in.length()-1)) {
		c = in.charAt(++i);
		if (c == 47) {

		} else if (c >= '0' && c <= '9') {
		    if (c != currentKey) {
			System.out.println("key " + c);
		    }
		    currentKey = c;
		} else if (c == '~') {
		    currentKey = 0;
		}
	    }
	    i++;
	}
    }

    public void printAsBytes( String s )
    {
	System.out.print("[");
	for (int i = 0; i < s.length(); i++) {
	    System.out.print("." + (int)s.charAt(i) + "(" + s.charAt(i) + ") ");
	}
	System.out.print("]\n");
    }

    public String getInputBuffer() {
	synchronized (inputBuffer) {
	    return inputBuffer.toString();
	}
    }

    public void clearInputBuffer() {
	synchronized (inputBuffer) {
	    inputBuffer = new StringBuffer();
	}
    }

    public String getEscInputBuffer() {
	synchronized (escInputBuffer) {
	    return escInputBuffer.toString();
	}
    }

    public void clearEscInputBuffer() {
	synchronized (escInputBuffer) {
	    escInputBuffer = new StringBuffer();
	}
    }

    public void serialEvent(SerialPortEvent e) {
	int newData = 0;

	// Determine type of event.
	switch (e.getEventType()) {
	    
	case SerialPortEvent.DATA_AVAILABLE:	
	    // Read data until -1 is returned. If \r is received substitute
	    // \n for correct newline handling.
	    synchronized (inputBuffer) {
		synchronized (escInputBuffer) {
		    while (newData != -1) {
			try {
			    newData = is.read();
			    
			    if (newData == -1) break;

			    if (expectingEscValue) {
				escInputBuffer.append((char) newData);
				expectingEscValue = false;
			    } else {
				if (DLE_BYTE == newData) {
				    expectingEscValue = true;
				} else if ('\r' == (char) newData) {
				    inputBuffer.append('\n');
				} else {
				    inputBuffer.append((char) newData);
				} 
			    }
			    
			} catch (IOException ex) {
			    System.err.println(ex);			
			    return;
			} 
		    }
		}
	    } 
	    
	    // Append received data to messageAreaIn.
	    //	    messageAreaIn.append(new String(inputBuffer));
	    //	    System.out.println("got data: " + inputBuffer);
	    
	    break;

	// If break event append BREAK RECEIVED message.
	case SerialPortEvent.BI:
	    System.out.println("\n--- BREAK RECEIVED ---\n");
	}
    }

    public static void main( String[] args ) 
	throws Exception
    {
	/*
	Enumeration e = CommPortIdentifier.getPortIdentifiers();

	while (e.hasMoreElements()) {
	    CommPortIdentifier id = (CommPortIdentifier)e.nextElement();

	    System.out.println("name=" + id.getName());
	}
	*/
	VoiceModemInterface vmi = new VoiceModemInterface();

	vmi.open();
	System.out.println("port open");
	
	vmi.initModem();
	System.out.println("modem initialized");

	System.out.println("waiting for ring...");
	vmi.waitForRing();
	System.out.println("got ring.");
	vmi.startTransmitMode();
	vmi.transmitFile("tmp/directions.rmd");
	vmi.endTransmitMode();

	vmi.close();
	System.out.println("closed");
	
    }
}
